/*
Copyright (c) 2009-2013, INRIA, Universite de Nancy 2 and Universidade
Federal do Rio Grande do Norte.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   * Neither the name of the Universite de Nancy 2 or the Universidade Federal
     do Rio Grande do Norte nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY INRIA, Universite de Nancy 2 and
Universidade Federal do Rio Grande do Norte ''AS IS'' AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INRIA, Université de Nancy 2 and
Universidade Federal do Rio Grande do Norte BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

/* DC number representation module */

#include "general.h"
#include "number.h"
#include "DAG-print.h"

static mpq_t mpq_zero;
static mpq_t mpq_one;

Tnumber
number_new()
{
  Tnumber n;
  MY_MALLOC(n, sizeof(struct TSnumber));
  mpq_init(n->num);
  return n;
}

Tnumber
number_dup(Tnumber src)
{
  Tnumber res;
  MY_MALLOC(res, sizeof(struct TSnumber));
  mpq_init(res->num);
  mpq_set(res->num, src->num);
  return res;
}

static Tnumber
number_set_si(Tnumber dest, const int num)
{
  mpq_set_si(dest->num, num, 1);
  return dest;
}

Tnumber
number_cpy(Tnumber dest, const Tnumber src)
{
  mpq_set(dest->num, src->num);
  return dest;
}

void 
number_free(Tnumber n)
{
  mpq_clear(n->num);
  free(n);
}

Tnumber
number_add(Tnumber res, const Tnumber n1, const Tnumber n2)
{
  mpq_add(res->num, n1->num, n2->num);
  
  return res;
}

Tnumber
number_sub(Tnumber res, const Tnumber n1, const Tnumber n2)
{
  mpq_sub(res->num, n1->num, n2->num);
  
  return res;
}

Tnumber
number_mul(Tnumber res, const Tnumber n1, const Tnumber n2)
{
  mpq_t t1;
  mpq_init(t1);
  mpq_mul(t1, n1->num, n2->num);
  mpq_set(res->num, t1);
  mpq_clear(t1);
  
  return res;
}

Tnumber
number_div(Tnumber res, const Tnumber n1, const Tnumber n2)
{
  mpq_t t1;
  mpq_init(t1);
  mpq_div(t1, n1->num, n2->num);
  mpq_set(res->num, t1);
  mpq_clear(t1);
  return res;
}

Tnumber
number_neg(Tnumber n)
{
  mpq_neg(n->num, n->num);
  
  return n;
}

/* DC Pre-condition: the double is a "int like", i.e., "x.00" */
Tnumber
number_from_double(Tnumber dest, const double d)
{
  if (d >= 0.0)
    mpq_set_si(dest->num, (int)(d+0.1), 1);
  else
    mpq_set_si(dest->num, (int)(d-0.1), 1);
  
  return dest;
}

void
number_from_DAG(Tnumber dest, const TDAG DAG)
{
  /* mpq_set_str does not handle dotted "decimal" literals." */
  if (mpq_set_str(dest->num, DAG_symb_name(DAG_symb(DAG)), 0) == -1)
    {
      /* this block deals with dotted decimal literals. */
      char * str = DAG_symb_name(DAG_symb(DAG)); /* str points to literal */
      char * tmp; /* point to a dot-less copy of str (to be computed) */
      size_t n; /* str = tmp \times 10^n (to be computed)*/
      size_t i;
      int status; /* used for sanity test */
      MY_MALLOC(tmp, strlen(str));
      i = 0;
      while (str[i] != '.' && str[i] != 0) /* copy integral part */
	{
	  tmp[i] = str[i];
	  ++i;
	}
      /* we must halt on a '.' otherwise mpq_set_str would have succeeded. */
      assert(str[i] == '.');
      ++i;
      n = 0;
      while(str[i] != 0) /* copy decimal part, compute divisor exponent n */
	{
	  tmp[i-1] = str[i];
	  ++i;
	  ++n;
	}
      --i;
      tmp[i] = 0; /* null-end target string */
      /* optimization: zeroes to the right are erased */
      --i;
      /* do not erase more than the decimal part */
      /* do not erase single, leftmost, zero (is ".000" possible?) */
      while(i > 0 && tmp[i] == '0' && n != 0) 
	{
	  tmp[i] = 0; /* maintain null-end */
	  --n;
	  --i;
	}
      status = mpq_set_str(dest->num, tmp, 10); /* must be base 10 */
      assert(status != -1);
      /* now divides by 10^n. */
      if (n != 0)
	{
	  mpz_t divisor_z;
	  mpq_t divisor_q;
	  mpz_init(divisor_z);
	  mpq_init(divisor_q);
	  mpz_ui_pow_ui(divisor_z, 10UL, (unsigned long) n);
	  mpq_set_z(divisor_q, divisor_z);
	  mpq_div(dest->num, dest->num, divisor_q);
	  mpz_clear(divisor_z);
	  mpq_clear(divisor_q);
	}
      free(tmp);
    }
  mpq_canonicalize(dest->num);
}

TDAG
number_to_DAG (const Tnumber number, Tsort sort)
{
  mpz_t den;
  mpz_t num;
  char * str;
  TDAG DAG;
  
  mpz_init (den);
  mpz_init (num);
  mpq_get_den (den, number->num);
  mpq_get_num (num, number->num);
  if (mpz_cmp_si(den, 1L) != 0)
    {
      str = mpq_get_str (NULL, 10, number->num);
      DAG = DAG_new_rational_str (str);
    }
  else if (sort == SORT_REAL)
    {
      char * str2;
      str = mpq_get_str(NULL, 10, number->num);
      MY_MALLOC(str2, (strlen(str)+3)*sizeof(char));
      sprintf(str2, "%s.0", str);
      DAG = DAG_new_decimal_str(str2);
      free(str2);
    }
  else
    {
      str = mpz_get_str (NULL, 10, num);
      DAG = DAG_new_numeral_str(str);
    }
  mpz_clear (den);
  mpz_clear (num);
  free (str);
  return DAG;
}

int
number_less(const Tnumber n1, const Tnumber n2)
{
  int cmp = mpq_cmp(n1->num, n2->num);
  return cmp < 0;
}

int
number_greater(const Tnumber n1, const Tnumber n2)
{
  int cmp = mpq_cmp(n1->num, n2->num);
  return cmp > 0;
}

/* DD the following code is *mostly* common to all implementations */

static int number_initialized = 0;

void
number_init()
{
  if (number_initialized)
    return;

  if (sizeof(int)*2 != sizeof(long long))
    my_error ("Fail in the number module due to the way the machine represents integers.\n");
     
  mpq_init (mpq_zero);
  mpq_init (mpq_one);
  mpq_set_si (mpq_one, 1, 1);

  number_zero = number_new();
  number_one = number_new();
  number_not_zero = number_new();

  number_set_si(number_zero, 0);
  number_set_si(number_one, 1);
  number_set_si(number_not_zero, 1);

  number_initialized = 1;
}

void
number_done()
{
  if (!number_initialized) 
    my_error ("Module number already closed.\n");

  number_free(number_zero);
  number_free(number_one);
  number_free(number_not_zero);

  mpq_clear (mpq_zero);
  mpq_clear (mpq_one);

  number_initialized = 0;
}

Tnumber
number_set(Tnumber dest, const int num)
{
  return number_set_si(dest, num);  
}

Tnumber 
number_abs(Tnumber n)
{
  if (number_less(n, number_zero))
    number_neg(n);
  return n;
}

