/*
Copyright (c) 2009-2013, INRIA, Universite de Nancy 2 and Universidade
Federal do Rio Grande do Norte.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   * Neither the name of the Universite de Nancy 2 or the Universidade Federal
     do Rio Grande do Norte nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY INRIA, Universite de Nancy 2 and
Universidade Federal do Rio Grande do Norte ''AS IS'' AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INRIA, Université de Nancy 2 and
Universidade Federal do Rio Grande do Norte BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

/*
  --------------------------------------------------------------
  arithmetic rewriting
  --------------------------------------------------------------
*/

#include <assert.h>
#include <stdlib.h>

#include "config.h"

#include "general.h"
#include "list.h"

#include "DAG.h"
#include "recursion.h"
#include "number.h"
#include "simp-arith.h"

#include "DAG-print.h"

/* IMPROVE: THIS SHOULD BE IN LA-pre.c */

/* ------------------------------------------------------------- */

static Tnumber num1; /** \brief auxiliary variable used in local routines */
static Tnumber num2; /** \brief auxiliary variable used in local routines */
static Tnumber num3; /** \brief auxiliary variable used in local routines */

/*
  --------------------------------------------------------------
  Module for preprocessing formulas aiming at Difference Logic
  --------------------------------------------------------------
*/

/**
   \brief checks DAG matches (X + ... + X) - (Y + ... + Y) - X, Y: wildcards
*/
#if 0
static int
DAG_match_diff_n_repeat(TDAG DAG)
{
  unsigned i;
  if (DAG_symb(DAG) != FUNCTION_MINUS)
    return 0;
  if (DAG_arity(DAG) != 2)
    return 0;
  if (DAG_symb(DAG_arg0(DAG)) != FUNCTION_SUM && DAG_symb(DAG_arg1(DAG)) != FUNCTION_SUM)
    return 0;
  if (DAG_arity(DAG_arg0(DAG)) != DAG_arity(DAG_arg1(DAG)))
    return 0;
  for (i = 1; i < DAG_arity(DAG_arg0(DAG_arg1(DAG))); i++)
    if (DAG_arg(DAG_arg0(DAG), i) != DAG_arg0(DAG_arg0(DAG)))
      return 0;
  for (i = 1; i < DAG_arity(DAG_arg1(DAG_arg1(DAG))); i++)
    if (DAG_arg(DAG_arg1(DAG), i) != DAG_arg0(DAG_arg1(DAG)))
      return 0;
  return 1;
}
#endif

/*
  --------------------------------------------------------------
  Rewrite t >= t' and t > t' to t' <= t and t' < t
  --------------------------------------------------------------
*/

static TDAG
greater_elim(TDAG src)
{
  if (DAG_symb(src) == PREDICATE_GREATER)
    return DAG_new_args(PREDICATE_LESS, DAG_arg1(src), DAG_arg(src, 0), DAG_NULL);
  else if (DAG_symb(src) == PREDICATE_GREATEREQ)
    return DAG_new_args(PREDICATE_LEQ, DAG_arg1(src), DAG_arg(src, 0), DAG_NULL);
  else
    return src;
}

/*
  --------------------------------------------------------------
  Rewrite (- c) and (/ c d) where c and d are numbers
  --------------------------------------------------------------
*/

static TDAG
rewrite_div(TDAG src)
{
  if (DAG_arity(src) != 2)
    my_error("rewrite_div: strange arity for src\n");
  if (DAG_is_number(DAG_arg0(src)) && DAG_is_number(DAG_arg1(src)))
    {
      TDAG dest = DAG_NULL;
      number_from_DAG(num1, DAG_arg0(src));
      number_from_DAG(num2, DAG_arg1(src));
      number_div(num3, num1, num2);
      dest = number_to_DAG(num3, DAG_sort(src));
      return dest;
    }
  return src;
}

/*--------------------------------------------------------------*/

static TDAG
rewrite_unary_minus(TDAG src)
{
  if (DAG_arity(src) != 1)
    my_error("rewrite_unary_minus: strange arity for src\n");
  if (DAG_is_number(DAG_arg(src, 0)))
    {
      TDAG dest = DAG_NULL;
      number_from_DAG(num1, DAG_arg(src, 0));
      number_neg(num1);
      dest = number_to_DAG(num1, DAG_sort(src));
      return dest;
    }
  return src;
}

/*--------------------------------------------------------------*/

static TDAG
rewrite_sum(TDAG src)
{
  unsigned i, j;
  unsigned numbers = 0, others = 0;
  TDAG * PDAG;
  if (DAG_arity(src) <= 1)
    my_error("rewrite_sum: strange arity for src\n");
  for (i = 0; i < DAG_arity(src); i++)
    if (DAG_is_number(DAG_arg(src, i)))
      {
        numbers++;
        if (numbers == 1)
          number_from_DAG(num1, DAG_arg(src, i));
        else
          {
            number_from_DAG(num2, DAG_arg(src, i));
            number_add(num1, num2, num1);
          }
      }
    else
      others++;
  assert(DAG_arity(src) == numbers+others);
  if (numbers == DAG_arity(src))
    return number_to_DAG(num1, DAG_sort(src));
  if (number_equal(num1, number_zero))
    {
      if (others == 1)
        {
          for (i = 0; i < DAG_arity(src); ++i)
            if (!DAG_is_number(DAG_arg(src, i)))
              return DAG_arg(src, i);
        }
      else
        {
          MY_MALLOC(PDAG, others * sizeof(*PDAG));
          for (i = j = 0; i < DAG_arity(src); ++i)
            if (!DAG_is_number(DAG_arg(src, i)))
              PDAG[j++] = DAG_arg(src, i);
          assert(j == others);
          return DAG_new(DAG_symb(src), others, PDAG);
        }
    }
  if (others == DAG_arity(src) || (numbers == 1 && DAG_is_number(DAG_argn(src))))
    return src;
  else
    {
      MY_MALLOC(PDAG, (others+1) * sizeof(*PDAG));
      for (i = j = 0; i < DAG_arity(src); ++i)
        if (!DAG_is_number(DAG_arg(src, i)))
          PDAG[j++] = DAG_arg(src, i);
      assert(j == others);
      PDAG[j] = number_to_DAG(num1, DAG_sort(src));
      return DAG_new(DAG_symb(src), others+1, PDAG);
    }
}

/*--------------------------------------------------------------*/

static TDAG
rewrite_minus(TDAG src)
{
  if (DAG_arity(src) != 2)
    my_error("rewrite_minus: strange arity for src\n");
  if (DAG_is_number(DAG_arg(src, 0)) && DAG_is_number(DAG_arg1(src)))
    {
      TDAG dest = DAG_NULL;
      number_from_DAG(num1, DAG_arg(src, 0));
      number_from_DAG(num2, DAG_arg1(src));
      number_sub(num3, num1, num2);
      dest = number_to_DAG(num3, DAG_sort(src));
      return dest;
    }
  return src;
}

/*--------------------------------------------------------------*/

static TDAG
rewrite_prod(TDAG src)
{
  unsigned i, j;
  unsigned numbers = 0, others = 0;
  TDAG * PDAG;
  if (DAG_arity(src) <= 1)
    my_error("rewrite_sum: strange arity for src\n");
  for (i = 0; i < DAG_arity(src); i++)
    if (DAG_is_number(DAG_arg(src, i)))
      {
        numbers++;
        if (numbers == 1)
          number_from_DAG(num1, DAG_arg(src, i));
        else
          {
            number_from_DAG(num2, DAG_arg(src, i));
            number_mul(num1, num2, num1);
          }
        if (number_equal(num1, number_zero))
          return DAG_ZERO;
      }
    else
      others++;
  assert(DAG_arity(src) == numbers+others);
  if (numbers == DAG_arity(src))
    return number_to_DAG(num1, DAG_sort(src));
  if (number_equal(num1, number_one))
    {
      if (others == 1)
        {
          for (i = 0; i < DAG_arity(src); ++i)
            if (!DAG_is_number(DAG_arg(src, i)))
              return DAG_arg(src, i);
        }
      else
        {
          MY_MALLOC(PDAG, others * sizeof(*PDAG));
          for (i = j = 0; i < DAG_arity(src); ++i)
            if (!DAG_is_number(DAG_arg(src, i)))
              PDAG[j++] = DAG_arg(src, i);
          assert(j == others);
          return DAG_new(DAG_symb(src), others, PDAG);
        }
    }
  if (others == DAG_arity(src) || (numbers == 1 && DAG_is_number(DAG_argn(src))))
    return src;
  else
    {
      MY_MALLOC(PDAG, (others+1) * sizeof(*PDAG));
      for (i = j = 0; i < DAG_arity(src); ++i)
        if (!DAG_is_number(DAG_arg(src, i)))
          PDAG[j++] = DAG_arg(src, i);
      assert(j == others);
      PDAG[j] = number_to_DAG(num1, DAG_sort(src));
      return DAG_new(DAG_symb(src), others+1, PDAG);
    }
}

/*--------------------------------------------------------------*/

/**
   \author Pascal Fontaine
   \brief rewrite arithmetic terms:
   unary minus applied to number is rewritten as number
   divisions of numbers are rewritten to the quotient
   sum including several numbers are rewritten to a sum with one number only
   number should be at the end
   number - number is rewritten to number
   \param src term (or formula)
   \return The rewritten term
   \remarks should be used inside structural recursion
*/
static TDAG
rewrite_node(TDAG src)
{
  if (unary_minus(DAG_symb(src)))
    return rewrite_unary_minus(src);
  if (DAG_symb(src) == FUNCTION_DIV)
    return rewrite_div(src);
  if (DAG_symb(src) == FUNCTION_SUM)
    return rewrite_sum(src);
  if (DAG_symb(src) == FUNCTION_MINUS)
    return rewrite_minus(src);
  if (DAG_symb(src) == FUNCTION_PROD)
    return rewrite_prod(src);
  return src;
}

/*
  --------------------------------------------------------------
  Public functions
  --------------------------------------------------------------
*/

static bool simp_arith_active = true;

/*--------------------------------------------------------------*/

void
simp_arith_logic(char * logic)
{
  if (strcmp(logic, "QF_UF") == 0)
    simp_arith_active = false;
}

/*--------------------------------------------------------------*/

TDAG
simp_arith(TDAG src)
{
  TDAG dest;
  if (!simp_arith_active)
    return src;
  dest = structural_recursion(src, greater_elim);
  DAG_free(src);
  src = dest;
  dest = structural_recursion(src, rewrite_node);
  DAG_free(src);
  return dest;
}

/*--------------------------------------------------------------*/

void
simp_arith_init(void)
{
  num1 = number_new();
  num2 = number_new();
  num3 = number_new();
}

/*--------------------------------------------------------------*/

void
simp_arith_done(void)
{
  number_free(num1);
  number_free(num2);
  number_free(num3);
}
