/*
Copyright (c) 2009-2013, INRIA, Universite de Nancy 2 and Universidade
Federal do Rio Grande do Norte.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   * Neither the name of the Universite de Nancy 2 or the Universidade Federal
     do Rio Grande do Norte nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY INRIA, Universite de Nancy 2 and
Universidade Federal do Rio Grande do Norte ''AS IS'' AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INRIA, Université de Nancy 2 and
Universidade Federal do Rio Grande do Norte BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

/**
  \file veriT-signal.c
  \author David Deharbe
  \brief Implementation of signal handling functionality. */

#define _XOPEN_SOURCE
#include <signal.h>
#include <unistd.h>
#include <stdio.h>

#include "general.h"
#include "stack.h"


#include "veriT-signal.h"

#ifndef	NSIG
#ifdef	_NSIG
#define	NSIG	_NSIG
#else
#define	NSIG	32
#endif
#endif

#define CAPTURE_SIGNALS
/*
   PF: there are several problems to solve here:
   - one cannot put a handler on SIGSTOP and SIGKILL, and thus I think
   this is not the right way to clear the system from eprover processes
   - I had plenty of problems on grid5k because of SIGSTOP signals while
   profiling.  I think there were generated by the profiler / not sure of it
*/

/*--------------------------------------------------------------*/

#ifdef CAPTURE_SIGNALS

TSstack(_sigfun, Tsigfun);

static Tstack_sigfun veriT_sigfun_stack;
#endif

/*--------------------------------------------------------------*/

#ifdef CAPTURE_SIGNALS
static void veriT_signal_handle(int signum)
{
  unsigned i;
  for (i = 0; i < stack_size(veriT_sigfun_stack); i ++)
    stack_get(veriT_sigfun_stack, i)();
  switch (signum)
    {
    case SIGILL:
      fprintf(stderr, "Illegal Instruction\n");
      break;
    case SIGFPE:
      fprintf(stderr, "Floating point exception\n");
      break;
    case SIGSEGV:
      fprintf(stderr, "Invalid memory reference\n");
      break;
    case SIGPIPE:
      fprintf(stderr, "Broken pipe\n");
      break;
    case SIGXCPU:
      fprintf(stderr, "CPU time limit exceeded\n");
      break;
    }

  _exit(signum);
}
#endif

/*--------------------------------------------------------------*/

static struct sigaction new_act, old;

void veriT_signal_init(void)
{
#ifdef CAPTURE_SIGNALS
  stack_INIT(veriT_sigfun_stack);
  memset(&new_act, 0, sizeof(new_act)); /* memset necessaire pour flags=0 */
  new_act.sa_handler = veriT_signal_handle;
  sigemptyset(&new_act.sa_mask);
  sigaddset(&new_act.sa_mask, SIGINT);
  sigaddset(&new_act.sa_mask, SIGQUIT);
  sigaddset(&new_act.sa_mask, SIGILL);
  sigaddset(&new_act.sa_mask, SIGABRT);
  sigaddset(&new_act.sa_mask, SIGFPE);
  sigaddset(&new_act.sa_mask, SIGSEGV);
  sigaddset(&new_act.sa_mask, SIGPIPE);
  sigaddset(&new_act.sa_mask, SIGTERM);
  sigaction(SIGINT, &new_act, &old);
  sigaction(SIGQUIT, &new_act, &old);
  sigaction(SIGILL, &new_act, &old);
  sigaction(SIGABRT, &new_act, &old);
  sigaction(SIGFPE, &new_act, &old);
  sigaction(SIGSEGV, &new_act, &old);
  sigaction(SIGPIPE, &new_act, &old);
  sigaction(SIGTERM, &new_act, &old);
  sigaction(SIGXCPU, &new_act, &old);
#endif
}

/*--------------------------------------------------------------*/

void veriT_signal_done(void)
{
#ifdef CAPTURE_SIGNALS
  new_act.sa_handler = SIG_DFL;  
  sigaction(SIGINT, &new_act, &old);
  sigaction(SIGQUIT, &new_act, &old);
  sigaction(SIGILL, &new_act, &old);
  sigaction(SIGABRT, &new_act, &old);
  sigaction(SIGFPE, &new_act, &old);
  sigaction(SIGSEGV, &new_act, &old);
  sigaction(SIGPIPE, &new_act, &old);
  sigaction(SIGTERM, &new_act, &old);
  sigaction(SIGXCPU, &new_act, &old);
  stack_free(veriT_sigfun_stack);
#endif
}

/*--------------------------------------------------------------*/

void veriT_signal_register(Tsigfun sigfun)
{
#ifdef CAPTURE_SIGNALS
  stack_push(veriT_sigfun_stack, sigfun);
#else
#ifdef PEDANTIC
  /* PF just to make sure arguments are used, so that no warning */
  printf("%p\n", sigfun);
#endif
#endif
}
